#include <iostream>
/* -*- c++ -*-
 *
 *  Signal class for inter-thread communication with argument support.
 *
 *  Copyright 2004 Christer Palm <palm@nogui.se>
 *
 *  Based on GtkmmDispatcher:
 *  Copyright 2000 Andreas Rottmannn <rottmann@users.sourceforge.net>
 *  Copyright 2004 Martin Schulze <mschulze@cvs.gnome.org>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Library General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Library General Public License for more details.
 *
 *  You should have received a copy of the GNU Library General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#include "dispatcher.h"

namespace GtkUtil {

namespace internal {

// DispatcherBase constructors
DispatcherBase::DispatcherBase() :
    disp_thread(Glib::Thread::self())
{
    // Make dispatch() invoke dispatch_self() from the dispatcher thread
    dispatch.connect(sigc::mem_fun(this, &DispatcherBase::dispatch_self));
}

DispatcherBase::DispatcherBase(const Glib::RefPtr<Glib::MainContext>& context) :
    disp_thread(Glib::Thread::self()),
    dispatch(context)
{
    DispatcherBase();
}

// DispatcherBase destructor
DispatcherBase::~DispatcherBase()
{
    // Make sure all outstanding signals are handled before destruction,
    // otherwise threads waiting for the completion of pending synchronous
    // signals might be left waiting forever
    tunnel(sigc::mem_fun(this, &DispatcherBase::null), true);
}

// Add a slot to the dispatch queue
void DispatcherBase::push(const sigc::slot<void>& slot)
{
    Glib::Mutex::Lock lock(queue_lock);
    queue.push_back(slot);
}

// Invoke all pending slots on dispatch queue
void DispatcherBase::dispatch_self()
{
    Glib::Mutex::Lock lock(queue_lock);
    while (!queue.empty()) {
        queue.front()();
        queue.pop_front();
    }
}

// Push a slot on the dispatch queue, invoke the dispatcher and
// optionally wait for the slot to complete execution
void DispatcherBase::tunnel(const sigc::slot<void>& slot, bool sync)
{
    push(slot);

    if (!sync) {
        dispatch();
    } else {
        if (Glib::Thread::self() == disp_thread) {
            // Avoid deadlock if caller thread is same as dispatcher thread
            dispatch_self();
        } else {
            Sync s;
            Glib::Mutex::Lock lock(s.mutex);
            // Push a slot on the dispatch queue that will call s.wakeup()
            // when invoked
            push(sigc::mem_fun(s, &Sync::wakeup));
            dispatch();
            // And wait for s.cond to be woken up
            s.cond.wait(s.mutex);
        }
    }
}

void DispatcherBase::null()
{
}

} /* namespace internal */

} /* namespace GtkUtil */
