/*
 * Glib::Dispatcher example -- cross thread signalling
 * by Daniel Elstner  <daniel.elstner@gmx.net>
 *
 * modified to only use glibmm
 * by J. Abelardo Gutierrez <jabelardo@cantv.net>
 *
 * modified for GtkUtil::Dispatcher
 * by Christer Palm <palm@nogui.se>
 *
 * Copyright (c) 2002-2003  Free Software Foundation
 */

#include <sigc++/class_slot.h>
#include <glibmm.h>

#include <iostream>
#include <algorithm>
#include <functional>
#include <list>

#include "dispatcher.h"

namespace
{
Glib::RefPtr<Glib::MainLoop> main_loop;

class ThreadProgress : public sigc::trackable
{
public:
  ThreadProgress(int id);
  virtual ~ThreadProgress();

  void launch();

  typedef GtkUtil::Dispatcher<ThreadProgress*> type_signal_finished;
  type_signal_finished& signal_finished();
  typedef GtkUtil::Dispatcher<ThreadProgress*, int> type_signal_progress;
  type_signal_progress& signal_progress();
  int id() const;

private:
  unsigned int        progress_;
  type_signal_progress signal_progress_;
  type_signal_finished signal_finished_;
  int                 id_;

  void progress_report(int id, int progress);
  void thread_function();
};

//TODO: Rename to avoid confusion with Glib::Dispatcher.
class Dispatcher : public sigc::trackable
{
public:
  Dispatcher();
  virtual ~Dispatcher();

  void launch_threads();

private:
  std::list<ThreadProgress*>  progress_list_;

  void on_progress_finished(ThreadProgress* progress);
  void on_progress(ThreadProgress* thread, int progress);
};


ThreadProgress::ThreadProgress(int id)
: 
  progress_ (0), id_ (id)
{
}

ThreadProgress::~ThreadProgress()
{}

void ThreadProgress::launch()
{
  // Create a non-joinable thread -- it's deleted automatically on thread exit.
  Glib::Thread::create(sigc::mem_fun(*this, &ThreadProgress::thread_function), false);
}

ThreadProgress::type_signal_finished& ThreadProgress::signal_finished()
{
  return signal_finished_;
}

ThreadProgress::type_signal_progress& ThreadProgress::signal_progress()
{
  return signal_progress_;
}

int ThreadProgress::id() const
{
  return id_;
}

void ThreadProgress::thread_function()
{
  Glib::Rand rand;
  int usecs = 5000;

  for(int i = 0; i < 100; ++i)
  {
    usecs = rand.get_int_range(std::max(0, usecs - 1000 - i), std::min(20000, usecs + 1000 + i));
    Glib::usleep(usecs);
    ++progress_;

    signal_progress().emit(this, progress_, true);
  }
  signal_finished().emit(this);
}

Dispatcher::Dispatcher()
{
  std::cout << "Thread Dispatcher Example." << std::endl;

  for(int i = 0; i < 5; ++i)
  {
    ThreadProgress *const progress = new ThreadProgress(i);
    progress_list_.push_back(progress);

    progress->signal_progress().connect(
        sigc::mem_fun(*this, &Dispatcher::on_progress));
    progress->signal_finished().connect(
        sigc::mem_fun(*this, &Dispatcher::on_progress_finished));
  }
}

Dispatcher::~Dispatcher()
{}

void Dispatcher::launch_threads()
{
  std::for_each(
      progress_list_.begin(), progress_list_.end(),
      std::mem_fun(&ThreadProgress::launch));
}

void Dispatcher::on_progress_finished(ThreadProgress* progress)
{
  std::cout << "Thread " << progress->id() << ": finished." << std::endl;

  progress_list_.remove(progress);

  if(progress_list_.empty())
    main_loop->quit();
}

void Dispatcher::on_progress(ThreadProgress* thread, int progress)
{
  std::cout << "Thread " << thread->id() << ": " << progress << " %" << std::endl;
}


} // anonymous namespace


int main(int argc, char** argv)
{
  Glib::thread_init();
  main_loop = Glib::MainLoop::create();

  Dispatcher dispatcher;

  // Install a one-shot idle handler to launch the threads
  Glib::signal_idle().connect(
      sigc::bind_return(sigc::mem_fun(dispatcher, &Dispatcher::launch_threads), false));

  main_loop->run();

  return 0;
}

